/*---------------------------------------------------------------------------------------------------
 Module:        C:\PalmDev\Projets\Mimee2\Geodesie.c
 Author:        Pierre Brial
 Project:       Geodesie
 State:         version 3
 Creation Date: 20 october 2005
 Description:   Geodesical functions set:

 ivf2e : ellipse flattening to excentricity
 geocentric : geographic to geocentric coordinates conversion
 geographic : geocentric to geographic coordinates conversion
 datumshift : datum1 to datum2 geocentric coordinates conversion
 LongueurMeridien : length of a meridien between Equator and given latitude
 ProjectionTM : Geographic to Transverse Mercator plane coordinates
 InvProjectionTM : Transverse Mercator plane to Geographic coordinates
 ProjectionConique : Geographic to Conic plane coordinates
 InvProjectionConique : Conic plane to Geographic coordinates
 ProjectionObliqueMercator : Geographical to Oblique Mercator coordinates
 InvProjectionObliqueMercator : Oblique Mercator to Geographic
 ProjectionCassini : Geographic to Cassini plane coordinates *Not Yet Tested*
 ProjectionStereographique : Geographical to Stereographic plane coordinates
 InvProjectionStereographique : Stereographic plane to geographical coordinates
 deg : degrees-minutes-seconds to decimal degree
 dms : decimal degree to degrees-minutes-seconds
 dmm2deg : degrees-decimal minutes to decimal degree
 deg2dmm : decimal degree to degrees-decimal minutes
 ZoneUTM : UTM zone given longitude
 LongitudeOrigineUTM : longitude of origin given UTM zone

Copyright :

     This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

---------------------------------------------------------------------------------------------------*/

#define radian 0.017453292519943
#define PI     3.141592653589793

struct coord{double x;double y;double z;};

struct ellipsoide
             {
             double a; // semi-major axis
             double e; // excentricity
             };

struct grille
             {
             // all angles in radian !
             double Lon0; // Longitude of origin
             double Lat0; // Latitude of origin
             double k;    // Scale factor
             double E0;   // False easting
             double N0;   // False northing
             double p1;   // Conic proj.: 1st standard parallel
                          // Oblique proj.: azimuth of the center line
                          //                passing through the centre
                          //                of the projection
             double p2;   // Conic proj.: 2nd standard parallel
                          // Oblique proj.: rectified bearing of the
                          //                center line (usually p1=p2)
             double p3;   // Conic proj.: correction on spherical longitude (rarely used)
                          // Oblique proj.: p3 = 1 : Hotine Oblique Mercator
                          //                p3 = 0 : Standard Oblique Merctor
             };

struct Datum
             {
             double dx; // geocentric X translation
             double dy; // geocentric Y translation
             double dz; // geocentric Z translation
             double d;  // correction of scale
             double rx; // rotation around X axis (radian)
             double ry; // rotation around Y axis (radian)
             double rz; // rotation around Z axis (radian)
             };

struct ConstOM{double B;double A;double H;double g0;double l0;double uc;};



/*------------------------------------------------------------------------
 Procedure:     ivf2e
 Purpose:       Compute ellipse's excentricity given inverse flattening
 Input:         ellipse's inverse flatening (1/f) (double)
 Output:        ellipse's excentricity (e) (double)
------------------------------------------------------------------------*/
double ivf2e(double ivf) {return sqrt((2-1/ivf)/ivf);}


/*------------------------------------------------------------------------
 Procedure:     geocentric
 Purpose:       geographical to geocentric coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                geographic coordinates in radians and meters (in structure coord)
 Output:        geocentric coordinates (in structure coord)
------------------------------------------------------------------------*/
struct coord geocentric(struct ellipsoide ellips,struct coord crd)
{
	struct coord crdout;
	double n=ellips.a/sqrt(1-ellips.e*ellips.e*pow(sin(crd.y),2));
	crdout.x=(n+crd.z)*cos(crd.y)*cos(crd.x);
	crdout.y=(n+crd.z)*cos(crd.y)*sin(crd.x);
	crdout.z=(n*(1-ellips.e*ellips.e)+crd.z)*sin(crd.y);
	return crdout;
}

/*------------------------------------------------------------------------
 Procedure:     geographic
 Purpose:       geocentric to geographical coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                geocentric coordinates (in structure coord)
 Output:        geographic coordinates in radians and meters(in structure coord)
------------------------------------------------------------------------*/
struct coord geographic(struct ellipsoide ellips,struct coord crd)
{
	struct coord crdout;
	double j,w,n,p0;
        double p = sqrt(crd.x*crd.x+crd.y*crd.y);
        crdout.y = atan(crd.z / p);
        crdout.x = atan(crd.y / crd.x);
        if (crd.x<0.0)
           {
           if (crd.y<0.0) crdout.x-=PI;
           else crdout.x+=PI;
           }
        boucle:
          {
          p0=crdout.y;
          j = sin(p0);
          w = sqrt(1 - ellips.e * ellips.e * j * j);
          n = ellips.a / w;
          crdout.y = atan(crd.z / p /
          (1 - n * ellips.e*ellips.e / (n + p * cos(p0) + crd.z * j - ellips.a * w)));
          }
        if (fabs(crdout.y-p0) > 0.000001) goto boucle;
        crdout.z = p / cos(crdout.y) - ellips.a / sqrt(1 - ellips.e * ellips.e * pow(sin(crdout.y),2));
        return crdout;
}

/*------------------------------------------------------------------------
 Procedure:     datumshift
 Purpose:       Convert geocentric coordinates from datum 1 to datum 2
 Input:         Datum 1 data, Datum 2 data (in structures datum),
                geocentric coordinates in datum 1 (in structure coord)
 Output:        geocentric coordinates in datum 2 (in structure coord)
------------------------------------------------------------------------*/
struct coord datumshift(struct Datum datum1,struct Datum datum2,struct coord crd)
{
       struct coord crdout;
       double delta_dx=datum1.dx-datum2.dx;
       double delta_dy=datum1.dy-datum2.dy;
       double delta_dz=datum1.dz-datum2.dz;
       double delta_d=datum1.d-datum2.d;
       double delta_rx=datum1.rx-datum2.rx;
       double delta_ry=datum1.ry-datum2.ry;
       double delta_rz=datum1.rz-datum2.rz;
       crdout.x= crd.x + delta_dx + delta_d * crd.x - delta_rz * crd.y + delta_ry * crd.z;
       crdout.y= crd.y + delta_dy + delta_d * crd.y + delta_rz * crd.x - delta_rx * crd.z;
       crdout.z= crd.z + delta_dz + delta_d * crd.z - delta_ry * crd.x + delta_rx * crd.y;
       return crdout;
}


/*-----------------------------------------------------------------------
 Procedure:     LongueurMeridien
 Purpose:       Meridian lenght between Equator and given latitude
 Input:         ellipsoide's data (in structure ellipsoide)
                Latitude (in radians)
 Output:        Meridian Lenght (in meters)
------------------------------------------------------------------------*/
double LongueurMeridien(struct ellipsoide ellips, double lat)
{
	double E2 = ellips.e * ellips.e;
	double E4 = E2*E2;
	double E6 = E4*E2;
	double E8 = E4*E4;
	double E10 = E8*E2;
	double B0 = 1+.75*E2+.703125*E4+.68359375*E6+.6729125977*E8+.6661834717*E10;
	double B2 = .375*E2+.46875*E4+.5013020833*E6+.5126953125*E8+.516027832*E10;
	double B4 = .1171875*E4+.205078125*E6+.259552002*E8+.2907943726*E10;
	double B6 = .04557291667*E6+.1025390625*E8+.1522705078*E10;
	double B8 = .01922607421875*E8+.0528717041015625*E10;
	double B10 = .00845947265625*E10;
	double x = cos(2 * lat);
	return ellips.a*(1-E2)*(B0*lat-sin(2*lat)*((((B10*x-B8)*x+B6)*x-B4)*x+B2));
}

/*------------------------------------------------------------------------
 Procedure:     ProjectionTM
 Purpose:       geographical to Transverse Mercator plane coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                geographic coordinates in radians and meters (in structure coord)
 Output:        plane coordinates in meters (in structure coord)
------------------------------------------------------------------------*/
struct coord ProjectionTM(struct ellipsoide ellips,struct grille gri,struct coord crd)
{
	struct coord crdout;
	double h = cos(crd.y);
	double B0 = LongueurMeridien(ellips,gri.Lat0);
	double B = LongueurMeridien(ellips,crd.y);
	double U = tan(crd.y);
	double T = U * U;
	double e2 = ellips.e*ellips.e;
	double EP = e2 / (1 - e2) * h * h;
	double R = ellips.a / sqrt(1 - pow((ellips.e * sin(crd.y)),2));
	double Z = (crd.x - gri.Lon0) * h;
	double z2 = Z * Z;
	double A3 = EP - T + 1;
	double A4 = EP * (4 * EP + 9) + 5 - T;
	double A5 = 2 * EP * (7 - 29 * T) + T * (T - 18) + 5;
	double A6 = 30 * EP * (9 - 11 * T) + T * (T - 58) + 61;
	crdout.x = gri.k * R * Z * (z2 * (A5 * z2 / 20 + A3) / 6 + 1) + gri.E0;
	crdout.y = gri.k * (B - B0 + R * U * z2 * (z2 * (A6 * z2 / 30 + A4) / 12 + 1) / 2) + gri.N0;
	crdout.z = crd.z;
	return crdout;
}

/*------------------------------------------------------------------------
 Procedure:     InvProjectionTM
 Purpose:       Transverse Mercator plane to geographical coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                plane coordinates in meters (in structure coord)
 Output:        geographic coordinates in radians and meters (in structure coord)
------------------------------------------------------------------------*/
struct coord InvProjectionTM(struct ellipsoide ellips,struct grille gri,struct coord crd)
{
       struct coord crdout;
       double EC = ellips.e * ellips.e;
       double ED = EC * EC;
       double EF = ED * EC;
       double EP = EC / (1 - EC);
       double x = crd.x-gri.E0;
       double m = (crd.y - gri.N0) / gri.k + LongueurMeridien (ellips,gri.Lat0);
       double U = m / ellips.a / (1 - EC / 4 - 0.046875 * ED - EF / 51.2);
       double ER = (1 - sqrt(1 - EC)) / (1 + sqrt(1 - EC));
       double P1 = (1.5 * ER - 0.84375 * pow(ER,3)) * sin(2 * U) + (1.3125 * ER * ER - 1.71875
                 * pow(ER,4)) * sin(4 * U)+ 151.0 * pow(ER,3) / 96.0 * sin(6 * U) + U;
       double G = 1 - EC * sin(P1) * sin(P1);
       double T = tan(P1);
       double h = cos(P1);
       double N1 = ellips.a / sqrt(G);
       double T1 = T * T;
       double C1 = EP * h * h;
       double R1 = ellips.a * (1 - EC) / pow(G,1.5);
       double D = (x / N1) / gri.k;
       double G1 = 5 + 3 * T1 + 10 * C1 - 4 * C1 * C1 - 9 * EP;
       double G2 = 61 + 90 * T1 + 298 * C1 + 45 * T1 * T1 - 252 * EP - 3 * C1 * C1;
       crdout.y = P1 - (N1 * T / R1 * (D * D / 2 - G1 * pow(D,4) / 24 + G2 * pow(D,6) / 720));
       crdout.x = (D - (1 + 2 * T1 + C1) * pow(D,3) / 6 + (5 - 2 * C1 + 28 * T1 - 3 * C1 * C1
                 + 8 * EP + 24 * T1 * T1) * pow(D,5) / 120) / h + gri.Lon0;
       crdout.z = crd.z;
       return crdout;
}

/*------------------------------------------------------------------------
 Procedure:     tc
 Purpose:       Function used by ProjectionConique and InvProjectionConique
------------------------------------------------------------------------*/
double tc(double e, double lat)
       {
       double ep=e*sin(lat);
       double t=tan(PI/4-lat/2)/pow((1-ep)/(1+ep),e/2);
	   if (t<0) t=0;
	   return t;
       }

/*------------------------------------------------------------------------
 Procedure:     ProjectionConique
 Purpose:       geographical to Conic plane coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                geographic coordinates in radians and meters (in structure coord)
 Output:        plane coordinates in meters (in structure coord)
------------------------------------------------------------------------*/
struct coord ProjectionConique(struct ellipsoide ellips,struct grille gri,struct coord crd)
       {
       struct coord crdout;
       double m1,m2,t1,t2,tF,t,n,F,r,rF,theta;

       if (gri.p1==gri.p2) gri.p1=gri.p2=gri.Lat0;
       else gri.k=1;

       m1=cos(gri.p1)/sqrt(1-pow(ellips.e*sin(gri.p1),2));
       m2=cos(gri.p2)/sqrt(1-pow(ellips.e*sin(gri.p2),2));
       t1=tc(ellips.e,gri.p1);
       t2=tc(ellips.e,gri.p2);
       tF=tc(ellips.e,gri.Lat0);
       t=tc(ellips.e,crd.y);

       if (gri.p1==gri.p2) n=sin(gri.p1);          // conic 1 Standard Parallel
       else n=(log(m1)-log(m2))/(log(t1)-log(t2)); // conic 2 Standard Parallel

       F=m1/(n*pow(t1,n));
       r=gri.k*ellips.a*F*pow(t,n);
       rF=gri.k*ellips.a*F*pow(tF,n);
       theta=n*(crd.x-gri.Lon0);

       crdout.x = gri.E0+r*sin(theta-gri.p3);
       crdout.y = gri.N0+rF-r*cos(theta-gri.p3);
       crdout.z = crd.z;
       return crdout;
       }

/*------------------------------------------------------------------------
 Procedure:     InvProjectionConique
 Purpose:       Conic plane to geographical coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                plane coordinates in meters (in structure coord)
 Output:        geographic coordinates in radians and meters (in structure coord)
------------------------------------------------------------------------*/
struct coord InvProjectionConique(struct ellipsoide ellips,struct grille gri,struct coord crd)
       {
       struct coord crdout;
       double n,F,rF,rp,tp,thetap,p0,ep,m1,m2,t1,t2,tF;
       if (gri.p1==gri.p2) gri.p1=gri.p2=gri.Lat0;
       else gri.k=1;
       m1=cos(gri.p1)/sqrt(1-pow(ellips.e*sin(gri.p1),2));
       m2=cos(gri.p2)/sqrt(1-pow(ellips.e*sin(gri.p2),2));
       t1=tc(ellips.e,gri.p1);
       t2=tc(ellips.e,gri.p2);
       tF=tc(ellips.e,gri.Lat0);
       if (gri.p1==gri.p2) n=sin(gri.p1);          // conic 1 Standard Parallel
       else n=(log(m1)-log(m2))/(log(t1)-log(t2)); // conic 2 Standard Parallel
       F=m1/(n*pow(t1,n));
       rF=gri.k*ellips.a*F*pow(tF,n);
       rp=sqrt(pow(crd.x-gri.E0,2)+pow(rF-crd.y+gri.N0,2));
       tp=pow(rp/(gri.k*ellips.a*F),1.0/n);
       thetap=atan((crd.x-gri.E0)/(rF-crd.y+gri.N0));
       crdout.x=(thetap+gri.p3)/n+gri.Lon0;
       crdout.y=gri.Lat0;
       do
         {
         p0=crdout.y;
         ep=ellips.e*sin(p0);
         crdout.y = PI/2-2*atan(tp*pow((1-ep)/(1+ep),ellips.e/2));
         }
       while (fabs(crdout.y-p0)>0.0000000001);
       crdout.z = crd.z;
       return crdout;
       }

/*------------------------------------------------------------------------
 Procedure:     ConstantesObliqueMercator
 Purpose:       Compute various constants used by ProjectionObliqueMercator
                and InvProjectionObliqueMercator
------------------------------------------------------------------------*/
struct ConstOM ConstantesObliqueMercator(struct ellipsoide ellips,struct grille gri)
	{
	struct ConstOM c;
	double D2,F,t0,D;
	double e2=pow(ellips.e,2);
	double esp=ellips.e*sin(gri.Lat0);
	double esp2=pow(esp,2);
	c.B=sqrt(1+e2*pow(cos(gri.Lat0),4)/(1-e2));
	c.A=ellips.a*c.B*gri.k*sqrt(1-e2)/(1-esp2);
	t0=tc(ellips.e,gri.Lat0);
	D=c.B*sqrt(1-e2)/cos(gri.Lat0)/sqrt(1-esp2);
	if (D<1) D2=1;
	else D2=D*D;
	F=D+copysign(sqrt(D2-1),gri.Lat0);
	c.H=F*pow(t0,c.B);
	c.g0=asin(sin(gri.p1)/D);
	c.l0=(F-1.0/F)/2.0*tan(c.g0);
	if (fabs(c.l0-1)<0.000000000001) c.l0=1;
	c.l0=gri.Lon0-asin(c.l0)/c.B;
	if (fabs(gri.p1-PI/2)<0.0000000001) c.uc=c.A*(gri.Lon0-c.l0);
	else c.uc=c.A/c.B*atan2(sqrt(D2-1),cos(gri.p1))*copysign(1,gri.Lat0);
	return c;
	}

/*------------------------------------------------------------------------
 Procedure:     ProjectionObliqueMercator
 Purpose:       geographical to Oblique Mercator plane coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                geographic coordinates in radians and meters (in structure coord)
 Output:        plane coordinates in meters (in structure coord)
------------------------------------------------------------------------*/
struct coord ProjectionObliqueMercator(struct ellipsoide ellips,struct grille gri,struct coord crd)
	{
	struct coord crdout;
	struct ConstOM c=ConstantesObliqueMercator(ellips,gri);
	double Q,S,U,V,v,u;

	Q=c.H/pow(tc(ellips.e,crd.y),c.B);
	S=(Q-1/Q)/2;
	V=sin(c.B*(crd.x-c.l0));
	U=(S*sin(c.g0)-V*cos(c.g0))/(Q+1/Q)*2;
	v=c.A*log((1-U)/(1+U))/2/c.B;
	u=c.A/c.B*atan2(S*cos(c.g0)+V*sin(c.g0),cos(c.B*(crd.x-c.l0)));//Hotine Oblique Mercator
	if (gri.p3==0) u-=c.uc;//Oblique Mercator
	crdout.x=v*cos(gri.p2)+u*sin(gri.p2)+gri.E0;
	crdout.y=u*cos(gri.p2)-v*sin(gri.p2)+gri.N0;
	crdout.z=crd.z;
	return crdout;
	}

/*------------------------------------------------------------------------
 Procedure:     InvProjectionObliqueMercator
 Purpose:       Oblique Mercator plane to geographical coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                plane coordinates in meters (in structure coord)
 Output:        geographic coordinates in radians and meters (in structure coord)
------------------------------------------------------------------------*/
struct coord InvProjectionObliqueMercator(struct ellipsoide ellips,struct grille gri,struct coord crd)
	{
	struct coord crdout;
	struct ConstOM c=ConstantesObliqueMercator(ellips,gri);
	double vp,up,Qp,Sp,Vp,Up,tp,x,e2,e4,e6,e8;

	vp=(crd.x-gri.E0)*cos(gri.p2)-(crd.y-gri.N0)*sin(gri.p2);
	up=(crd.y-gri.N0)*cos(gri.p2)+(crd.x-gri.E0)*sin(gri.p2);
	if (gri.p3==0) up+=c.uc; //Oblique Mercator
	Qp=exp(-c.B*vp/c.A);
	Sp=(Qp-1/Qp)/2;
	Vp=sin(c.B*up/c.A);
	Up=(Vp*cos(c.g0)+Sp*sin(c.g0))/(Qp+1/Qp)*2;

	tp=pow(c.H/sqrt((1+Up)/(1-Up)),1/c.B);

	x=PI/2-2*atan(tp);
	e2=ellips.e*ellips.e;
	e4=e2*e2;
	e6=e4*e2;
	e8=e4*e4;
	crdout.y=x+sin(2*x)*(e2/2+5*e4/24+e6/12+13*e8/360)
				+sin(4*x)*(7*e4/48+29*e6/240+811*e8/11520)
				+sin(6*x)*(7*e6/120+81*e8/1120)
				+sin(8*x)*(4279*e8/161280);

	crdout.x=c.l0-atan2(Sp*cos(c.g0)-Vp*sin(c.g0),cos(c.B*up/c.A))/c.B;

	crdout.z=crd.z;
	return crdout;
	}


/*------------------------------------------------------------------------
 Procedure:     ProjectionCassini
 Warning :      *** Not Tested !!!***
 Purpose:       geographical to Cassini plane coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                geographic coordinates in radians and meters (in structure coord)
 Output:        plane coordinates in meters (in structure coord)
------------------------------------------------------------------------*/
/*
struct coord ProjectionCassini(struct ellipsoide ellips,struct grille gri,struct coord crd)
{
	struct coord crdout;
	double h = cos(crd.y);
	double B0=LongueurMeridien(ellips,gri.Lat0);
	double B=LongueurMeridien(ellips,crd.y);
	double U = tan(crd.y);
	double T = U * U;
	double e2=ellips.e * ellips.e;
	double EP = e2 / (1 - e2) * h * h;
	double R = ellips.a / sqrt(1 - pow((ellips.e * sin(crd.y)),2));
	double Z = (crd.x - gri.Lon0) * h;
	double z2 = Z * Z;
	crdout.x = R * (Z-T*pow(Z,3)/6 - (8-T+8*EP) * T* pow(Z,5)/120) + gri.E0;
	crdout.y = B - B0 + R * U * (z2/2 + (5-T+6*EP)*pow(z2,2)/24) + gri.N0;
	crdout.z=crd.z;
	return crdout;
}
*/


/*------------------------------------------------------------------------
 Procedure:     ProjectionStereographique
 Purpose:       geographical to Stereographic plane coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                geographic coordinates in radians and meters (in structure coord)
 Output:        plane coordinates in meters (in structure coord)
------------------------------------------------------------------------*/
struct coord ProjectionStereographique(struct ellipsoide ellips,struct grille gri,struct coord crd)
	{
	struct coord crdout;
	double e2 = ellips.e*ellips.e;
	double r0=ellips.a*(1-e2)/pow(1 - e2*pow(sin(gri.Lat0),2),1.5);
	double n0=ellips.a / sqrt(1 - pow((ellips.e * sin(gri.Lat0)),2));
	double R=sqrt(r0*n0);
	double n=sqrt(1+e2*pow(cos(gri.Lat0),4)/(1-e2));
	double sinp0=sin(gri.Lat0);
	double S1=(1+sinp0)/(1-sinp0);
	double S2=(1-ellips.e*sinp0)/(1+ellips.e*sinp0);
	double w1=pow(S1*pow(S2,ellips.e),n);
	double sinx0=(w1-1)/(w1+1);
	double c=(n+sinp0)*(1-sinx0)/(n-sinp0)/(1+sinx0);
	double w2=c*w1;
	double x0=asin((w2-1)/(w2+1));
	double Lambda=n*(crd.x-gri.Lon0)+gri.Lon0;
	double sinp=sin(crd.y);
	double Sa=(1+sinp)/(1-sinp);
	double Sb=(1-ellips.e*sinp)/(1+ellips.e*sinp);
	double w=c*pow(Sa*pow(Sb,ellips.e),n);
	double x=asin((w-1)/(w+1));
	double beta=1+sin(x)*sin(x0)+cos(x)*cos(x0)*cos(Lambda-gri.Lon0);
	crdout.x = gri.E0+2*R*gri.k*cos(x)*sin(Lambda-gri.Lon0)/beta;
	crdout.y = gri.N0+2*R*gri.k*(sin(x)*cos(x0)-cos(x)*sin(x0)*cos(Lambda-gri.Lon0))/beta;
	crdout.z=crd.z;
	return crdout;
	}

/*------------------------------------------------------------------------
 Procedure:     InvProjectionStereographique
 Purpose:       Stereographic plane to geographical coordinates conversion
 Input:         ellipsoid's data (in structure ellipsoide)
                grid's data (in structure grille)
                plane coordinates in meters (in structure coord)
 Output:        geographic coordinates in radians and meters (in structure coord)
------------------------------------------------------------------------*/
struct coord InvProjectionStereographique(struct ellipsoide ellips,struct grille gri,struct coord crd)
	{
	struct coord crdout;
	double e2 = ellips.e*ellips.e;
	double r0=ellips.a*(1-e2)/pow(1 - e2*pow(sin(gri.Lat0),2),1.5);
	double n0=ellips.a / sqrt(1 - pow((ellips.e * sin(gri.Lat0)),2));
	double R=sqrt(r0*n0);
	double n=sqrt(1+e2*pow(cos(gri.Lat0),4)/(1-e2));
	double sinp0=sin(gri.Lat0);
	double S1=(1+sinp0)/(1-sinp0);
	double S2=(1-ellips.e*sinp0)/(1+ellips.e*sinp0);
	double w1=pow(S1*pow(S2,ellips.e),n);
	double sinx0=(w1-1)/(w1+1);
	double c=(n+sinp0)*(1-sinx0)/(n-sinp0)/(1+sinx0);
	double w2=c*w1;
	double X0=asin((w2-1)/(w2+1));
	double g=2*R*gri.k*tan(M_PI/4-X0/2);
	double h=4*R*gri.k*tan(X0)+g;
	double i=atan((crd.x-gri.E0)/(h+crd.y-gri.N0));
	double j=atan((crd.x-gri.E0)/(g-crd.y+gri.N0))-i;
	//printf("%.3f\t%.3f\t%.9f\t%.9f\n",g,h,i,j);
	double X=X0+2*atan((crd.y-gri.N0-(crd.x-gri.E0)*tan(j/2))/(2*R*gri.k));
	//printf("X = %.8f\n",X);
	double Lambda=j+2*i+gri.Lon0;
	//printf("Lambda = %.8f\n",Lambda);
	double psi=0.5*log((1+sin(X))/c/(1-sin(X)))/n;
	double p0,psi0,psi1;
	crdout.y=2*atan(exp(psi))-M_PI/2;
	psi1=log((tan(crdout.y/2+M_PI/4))*pow((1-ellips.e*sin(crdout.y))/(1+ellips.e*sin(crdout.y)),ellips.e/2));
	//printf("psi = %.8f\tphi = %.8f\n",psi,crdout.y);
	do
         {
         p0=crdout.y;
         psi0=psi1;
         crdout.y=p0-(psi1-psi)*cos(p0)*(1-e2*pow(sin(p0),2))/(1-e2);
         psi1=log((tan(crdout.y/2+M_PI/4))*pow((1-ellips.e*sin(crdout.y))/(1+ellips.e*sin(crdout.y)),ellips.e/2));
         //printf("%.10f\n",crdout.y-p0);
         }
       while (fabs(crdout.y-p0)>0.0000000001);
       //while (fabs(crdout.y-p0)>0.00001);
    crdout.x=(Lambda-gri.Lon0)/n+gri.Lon0;
	crdout.z=crd.z;
	return crdout;
	}


/*** Various Sexagesimal conversion functions ***/

double deg(double x) // D.MMSSss to D.DDD
{
	double ax,s,h,m;
	ax = fabs(x) + 1e-12;
	h = floor(ax);
	m = floor((ax - h) * 100);
	s = (ax - h) * 10000 - m * 100;
	return copysign(h+m/60+s/3600,x);
}

double dms(double x) // D.DDD to D.MM.SSss
{
	double ax,h,m,s;
        ax = fabs(x) + 1e-12;
	h = floor(ax);
	m = (ax - h) * 60;
	s = (m - floor(m)) * 0.006;

	return copysign(h + floor(m) / 100 + s,x);
}

double dmm2deg(double x) // D.MMmm to D.DDD
{
	double d = copysign(floor(abs(x)),x);
	return d + (x - d) / 0.6;
}

double deg2dmm(double x) // D.DDD to D.MMmm
{
	double d = copysign(floor(abs(x)),x);
	return d + (x - d) * 0.6;
}


/*** UTM Zones functions ***/

int ZoneUTM(double longitude) // Compute UTM Zone number given longitude in degree
{return floor((longitude/radian + 180) / 6) + 1;}


double LongitudeOrigineUTM(int zone) // compute UTM zone longitude of origin (in radian)
                                     // given UTM zone number
{return (double)((zone - 1) * 6 - 177)*radian;}

